__id__ = "$Id: xtuple.py 572 2007-10-22 23:53:25Z theobald $"

class Lineage:
    """Represent a lineage of an alternative in Trio.
    """
   
    def __init__( self ):
        """Constructor.""" 
        self.srctable = []
        self.srcaid = []
        self.disjlin = [] #delta of 0 means conjunctive, 1 means disj, 2 means "and not"
        
    def append(self, table, aid, disjlin=0):
        self.srctable.append(table)
        self.srcaid.append(aid)
        self.disjlin.append(disjlin)
    
    def tables(self):
      return self.srctable
    
    def aids(self): 
      return self.srcaid
      
    def disj(self): 
      return self.disjlin
    
    def items(self):
        return map(None, self.srctable, self.srcaid, self.disjlin)
      
    def aidstrs(self): 
      return [str(x) for x in self.srcaid]
      
class Alternative:
    """Represent an alternative in Trio.

    @ivar data: Attribute values for this alternative. 
    @type data: A list. 
    @ivar lineage: Lineage of this alternative.
    @type lineage: A L{Lineage} object.
    @ivar confidence: Confidence value for this alternative.
    @type confidence: float.
    """
   
    def __init__(self, data, lineage=Lineage(), conf=None):
        """Constructor for Alternative objects"""
        self.data = data
        self.lineage = lineage
        self.confidence = conf
      
    def __getitem__(self, index):
        """Return an attribute at the location specified by index."""
        return self.data[index]

    def __delitem__(self, index):
        """delete atem"""
        del self.data[index]
      
    def __len__(self):
        """Return number of attributes in this alternative."""
        return len(self.data)

    def __str__(self):
        result = "(%s)" % ",".join(str(x) for x in self.data)
        if self.confidence is not None:
            result += ":%.3f" % self.confidence
        return result

    def append(self, attr):
        """Append an attribute value to the alternative"""
        self.data.append(attr)

    def getLineage(self):
        return self.lineage
        
    def getConfidence(self):
        return self.confidence   

    def traceLineage(self, tcnx):
        tnc = tcnx.ncursor()
        lin = self.lineage.items()
        # transient result
        if len(lin) > 1:
            return lin
        # persistent result
        elif len(lin) == 1:
            query = """select srctable, srclin, disjlin from trio_lin_%s where aid = %s order by disjlin asc""" % (lin[0][0], lin[0][1])
            tnc.execute(query)
            return tnc.fetchall()
        else:
            return []
    
    def computeConfidence(self, tcnx):
        tnc = tcnx.ncursor()
        tables = ",".join("'%s'" % x  for x in self.lineage.tables())
        aids = ",".join("%s" % str(x) for x in self.lineage.aids())
        disj = ",".join("%s" % str(x) for x in self.lineage.disj())
        query = "select trio_get_conf(array[" +  \
            tables + "],array[" +  \
            aids + "],array[" + disj + "],NULL,0,1)";
        conf = None
        try:
            tnc.execute(query)
            conf = tnc.fetchone()[0]
        except Exception, ex:
            tcnx.rollback()
            conf = "error"
        return conf

    def setConfidence(self, conf):
        self.confidence = conf 

class XTuple:
    def __init__(self):
        """A constructor for XTuple."""
        self.alternatives = []
        self.enc = 0

    def __str__(self):
        alts = [str(x) for x in self.alternatives]
        alts.sort()
        return "|".join(alts) + "\n"

    def __getitem__(self, index):
        """Returns the alternative at the specified index."""
        return self.alternatives[index]

    def __len__(self):
        """Return number of alternatives in this object."""
        return len(self.alternatives)

    def getAlternative(self, index):
        """Returns the alternative at the specified index."""
        return self.alternatives[index]
        
    def append(self, alternative):
        """Append alternative to this xtuple.
        
        @param alternative: Alternative to be appended to this xtuple. 
        @type alternative: L{Alternative}
        """ 
        self.alternatives.append(alternative)
        self.enc += 1
      
    def getConfidence(self):
        """Return sum of confidence values."""
        if self.alternatives[0].getConfidence() is None:
            return None
        try:
            num = sum(x.getConfidence() for x in self.alternatives)
        except:
            num = 0.0
        if num < 0.0:
            return None
        else:
            return num

    def getQuestionMark(self):
        if self.getConfidence() is not None:
            return self.getConfidence() < 1.0
        else:
            return self.enc > len(self.alternatives)

    def setQuestionMark(self):
        self.enc += 1
